from leapp.workflows.api.v1.depcheck import DepCheckAPI1, DepCheckAPI2, DepCheckAPI3, DepCheckAPI4
from leapp.workflows.api.v2.newapi import NewAPI
from leapp.workflows.api.v2.testapi import TestAPI


__all__ = (
    DepCheckAPI1,
    DepCheckAPI2,
    DepCheckAPI3,
    DepCheckAPI4,
    TestAPI,
    NewAPI
)
