from leapp.utils.clicmd import command

_LONG_DESCRIPTION = '''
Leapp Workflow related commands.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('workflow', help='Workflow related commands', description=_LONG_DESCRIPTION)
def workflow(*args):  # noqa; pylint: disable=unused-argument
    pass
